import { Argv, Arguments } from 'yargs'
import { errormsg } from '../../../helpers'
import { getProviderByContext } from '../../../index'

export const command: string = 'file'
export const desc: string = 'Fetch a raw file from a git repository'

export const builder = (yargs: Argv) => yargs
    .usage('Usage: $0 get file [options]')
    .option('owner', {
        alias: 'o',
        demandOption: errormsg`must provide "owner" option`,
        describe: 'The name of the owner of the git repository',
        type: 'string'
    })
    .option('repo', {
        alias: 'r',
        demandOption: errormsg`must provide "repo" option`,
        describe: 'The name of the git repository',
        type: 'string'
    })
    .option('revision', {
        alias: ['ref', 'v'],
        describe: 'The revision of the file',
        type: 'string',
        default: 'master'
    })
    .option('path', {
        alias: 'p',
        demandOption: errormsg`must provide "path" option`,
        describe: 'The path to the file that needs to be fetched',
        type: 'string'
    })
    .example(
        `$0 get ${command} -o some-owner -r some-repo -p ./path/to/file`,
        'Will fetch the file located at path/to/file from the root of repo: "some-owner/some-repo" (revision "master")'
    )
    .example(
        `$0 get ${command} -o some-owner -r some-repo -p ./path/to/file -v some-rev `,
        'Will fetch the file located at path/to/file from the root of repo: "some-owner/some-repo" revision "some-rev"'
    )
    .example(
        `$0 get ${command} -o some-owner -r some-repo -p ./path/to/file -s output.txt`,
        'Will fetch the file located at path/to/file from the root of repo: "some-owner/some-repo" and save the contents to "output.txt"'
    )

export const handler = async (argv: Arguments) => {
    const owner = argv.owner as string
    const repo = argv.repo as string
    const path = argv.path as string
    const revision = argv.revision as string
    const gitContext = argv.gitContext

    console.log(`owner: ${argv.owner}`)
    console.log(`repo: ${argv.repo}`)
    console.log(`revision: ${argv.revision}`)
    console.log(`path: ${argv.path}`)
    const provider = await getProviderByContext(gitContext)
    const file = await provider.fetchRawFile({ owner, repo, ref: revision, path })
    console.log(file)
}