jest.mock('request-promise');

const request = require('request-promise');
const BitbucketServer = require('../bitbucket-server').default;
const { to } = require('../../helpers');

const getBranchResponse = (status = 200, lastPage = true, branch = {}) => {
    return {
        statusCode: status,
        body: {
            values: [
                { id: 'refs/heads/master',
                    displayId: 'master',
                    latestCommit: 'some-hash',
                    metadata: {
                        'com.atlassian.bitbucket.server.bitbucket-branch:latest-commit-metadata': {
                            id: 'some-hash',
                            author: {
                                name: 'joe schmo',
                                emailAddress: 'joe@gmail.com',
                            },
                            message: 'some message',
                        }
                    }, ...branch },
            ],
            isLastPage: lastPage,
        },
    };
};

const getFileResponse = (status = 200, lastPage = true) => {
    return {
        statusCode: status,
        body: {
            lines: [
                {
                    text: 'some text',
                },
                {
                    text: 'some more text',
                },
                {
                    text: 'and some more text',
                }
            ],
            isLastPage: lastPage,
        },
    };
};

describe('Bitbucket-server git provider tests', () => {
    describe('Constructor', () => {
        it('Should construct a client', () => {
            const client = new BitbucketServer({ username: 'uzer', password: '1234abcd', type: 'basic', apiURL: 'http://bitbucket.server/api' });
            expect(client).toBeInstanceOf(BitbucketServer);
            expect(client.fetchRawFile).toBeDefined();
            expect(client.getBranch).toBeDefined();
        });

        it('Should use the provided token in the Authorization header', async () => {
            const client = new BitbucketServer({ username: 'uzer', password: '1234abcd', type: 'basic', apiURL: 'http://bitbucket.server/api' });
            const reqMock = jest.fn(async () => getFileResponse());
            request.mockImplementationOnce(reqMock);

            await client.fetchRawFile({ owner: 'joe', repo: 'some-repo', ref: 'master', path: 'some-path' });
            expect(reqMock).toHaveBeenCalledTimes(1);
            expect(reqMock.mock.calls[0][0].headers).toEqual({ Authorization: `Basic ${Buffer.from('uzer:1234abcd').toString('base64')}` });
        });

        it('Should fail if api url not provided (no default url)', async () => {
            const [err] = await to(async () => new BitbucketServer({ username: 'uzer', password: '1234abcd', type: 'basic' }));

            expect(err).toBeDefined();
            expect(err.message).toEqual('Cannot create bitbucket-server provider without api url');
        });

        it('Should use the provided api url', async () => {
            const client = new BitbucketServer({ username: 'uzer', password: '1234abcd', type: 'basic', apiURL: 'http://bitbucket.server' });
            const reqMock = jest.fn(async () => getFileResponse());
            request.mockImplementationOnce(reqMock);

            await client.fetchRawFile({ owner: 'joe', repo: 'some-repo', ref: 'master', path: 'some-path' });
            expect(reqMock).toHaveBeenCalledTimes(1);
            expect(reqMock.mock.calls[0][0].url).toEqual('http://bitbucket.server/rest/api/1.0/projects/joe/repos/some-repo/browse/some-path');
        });
    });

    describe('Positive', () => {
        describe('fetchRawFile', () => {
            it('Should fetch the requested file in one page', async () => {
                const client = new BitbucketServer({ username: 'uzer', password: '1234abcd', type: 'basic', apiURL: 'http://bitbucket.server' });
                const reqMock = jest.fn(async () => getFileResponse());
                request.mockImplementationOnce(reqMock);

                const res = await client.fetchRawFile({ owner: 'joe', repo: 'some-repo', ref: 'master', path: 'some-path' });
                expect(reqMock).toHaveBeenCalledTimes(1);
                expect(res).toEqual('some text\nsome more text\nand some more text');
            });

            it('Should fetch the requested file in multiple pages', async () => {
                const client = new BitbucketServer({ username: 'uzer', password: '1234abcd', type: 'basic', apiURL: 'http://bitbucket.server' });
                const reqMock = jest.fn();
                request.mockImplementationOnce(reqMock);
                request.mockImplementationOnce(reqMock);
                reqMock.mockImplementationOnce(async () => getFileResponse(200, false));
                reqMock.mockImplementationOnce(async () => getFileResponse(200, true));

                const res = await client.fetchRawFile({ owner: 'joe', repo: 'some-repo', ref: 'master', path: 'some-path' });
                expect(reqMock).toHaveBeenCalledTimes(2);
                expect(res).toEqual('some text\nsome more text\nand some more text\nsome text\nsome more text\nand some more text');
            });
        });

        describe('getBranch', () => {
            it('Should get the requested branch', async () => {
                const client = new BitbucketServer({ username: 'uzer', password: '1234abcd', type: 'basic', apiURL: 'http://bitbucket.server' });
                const reqMock = jest.fn(async () => getBranchResponse());
                request.mockImplementationOnce(reqMock);

                const res = await client.getBranch({ owner: 'joe', repo: 'some-repo', branch: 'master' });
                const branchResponse = getBranchResponse();
                expect(reqMock).toHaveBeenCalledTimes(1);
                expect(reqMock.mock.calls[0][0].qs.filterText).toEqual('master');
                expect(reqMock.mock.calls[0][0].qs.details).toEqual('true');
                expect(res.id).toEqual(branchResponse.body.values[0].id);
                expect(res.name).toEqual(branchResponse.body.values[0].displayId);
                const commit = branchResponse.body.values[0].metadata['com.atlassian.bitbucket.server.bitbucket-branch:latest-commit-metadata'];
                expect(res.commit.commiter_name).toEqual(commit.author.name);
                expect(res.commit.message).toEqual(commit.message);
                expect(res.commit.sha).toEqual(commit.id);
                expect(res.commit.url).toEqual('');
            });

            it('Should get the requested branch in multiple pages if it has no displayId', async () => {
                const client = new BitbucketServer({ username: 'uzer', password: '1234abcd', type: 'basic', apiURL: 'http://bitbucket.server' });
                const reqMock = jest.fn();
                request.mockImplementationOnce(reqMock);
                request.mockImplementationOnce(reqMock);
                reqMock.mockImplementationOnce(async () => getBranchResponse(200, false));
                const branchResponse = getBranchResponse(200, true, { displayId: undefined, id: 'refs/heads/some-branch' });
                reqMock.mockImplementationOnce(async () => branchResponse);

                const res = await client.getBranch({ owner: 'joe', repo: 'some-repo', branch: 'some-branch' });

                expect(reqMock).toHaveBeenCalledTimes(2);
                expect(reqMock.mock.calls[0][0].qs.filterText).toEqual('some-branch');
                expect(reqMock.mock.calls[0][0].qs.details).toEqual('true');
                expect(res.id).toEqual(branchResponse.body.values[0].id);
                expect(res.name).toEqual(branchResponse.body.values[0].displayId);
                const commit = branchResponse.body.values[0].metadata['com.atlassian.bitbucket.server.bitbucket-branch:latest-commit-metadata'];
                expect(res.commit.commiter_name).toEqual(commit.author.name);
                expect(res.commit.message).toEqual(commit.message);
                expect(res.commit.sha).toEqual(commit.id);
                expect(res.commit.url).toEqual('');
            });
        });
    });

    describe('Negative', () => {
        describe('fetchRawFile', () => {
            it('Should fail if the requested file is not found', async () => {
                const client = new BitbucketServer({ username: 'uzer', password: '1234abcd', type: 'basic', apiURL: 'http://bitbucket.server' });
                const reqMock = jest.fn(async () => getFileResponse(404));
                request.mockImplementationOnce(reqMock);

                const opts = { owner: 'joe', repo: 'some-repo', ref: 'master', path: 'some-path' };
                const [err] = await to(client.fetchRawFile(opts));

                expect(err).toBeDefined();
                expect(err.message).toEqual(`Failed to retrieve file: ${JSON.stringify(opts)}, status code: 404`);
            });
        });

        describe('getBranch', () => {
            it('Should fail if the requested branch is not found', async () => {
                const client = new BitbucketServer({ username: 'uzer', password: '1234abcd', type: 'basic', apiURL: 'http://bitbucket.server' });
                const reqMock = jest.fn(async () => getBranchResponse(404));
                request.mockImplementationOnce(reqMock);

                const opts = { owner: 'joe', repo: 'some-repo', branch: 'master' };
                const [err] = await to(client.getBranch(opts));

                expect(err).toBeDefined();
                expect(err.message).toEqual(`Failed to retrieve branch: ${JSON.stringify(opts)}, status code: 404`);
            });
        });
    });
});
