jest.mock('request-promise');

const request = require('request-promise');
const Bitbucket = require('../bitbucket').default;
const { to } = require('../../helpers');

const getBranchResponse = (status = 200) => {
    return {
        statusCode: status,
        body: {
            name: 'master',
            target: {
                hash: '05a7d037f1317877cf1719db1c231a08cc6c7788',
                message: 'some commit message',
                author: {
                    user: {
                        display_name: 'Joe Schmo',
                    },
                },
                links: {
                    html: {
                        href: 'https://bitbucket.org/!api/2.0/repositories/roikramer120/test-repo/patch/2c22488663e2619f524f8adc729886ba921f5fdb'
                    }
                }
            },
        },
    };
};

const getFileResponse = (status = 200) => {
    return {
        statusCode: status,
        body: 'some-text',
    };
};

describe('Bitbucket git provider tests', () => {
    describe('Constructor', () => {
        it('Should construct a client', () => {
            const client = new Bitbucket({});
            expect(client).toBeInstanceOf(Bitbucket);
            expect(client.fetchRawFile).toBeDefined();
            expect(client.getBranch).toBeDefined();
        });

        it('Should use the provided token in the Authorization header', async () => {
            const client = new Bitbucket({ password: '1234abcd', type: 'oauth2' });
            const reqMock = jest.fn(async () => getFileResponse());
            request.mockImplementationOnce(reqMock);

            await client.fetchRawFile({ owner: 'joe', repo: 'some-repo', ref: 'master', path: 'some-path' });
            expect(reqMock).toHaveBeenCalledTimes(1);
            expect(reqMock.mock.calls[0][0].headers).toEqual({ Authorization: 'Bearer 1234abcd' });
        });

        it('Should use the provided username and password in the Authorization header', async () => {
            const client = new Bitbucket({ username: 'uzer', password: '1234abcd', type: 'basic' });
            const reqMock = jest.fn(async () => getFileResponse());
            request.mockImplementationOnce(reqMock);

            await client.fetchRawFile({ owner: 'joe', repo: 'some-repo', ref: 'master', path: 'some-path' });
            expect(reqMock).toHaveBeenCalledTimes(1);
            expect(reqMock.mock.calls[0][0].headers).toEqual({ Authorization: `Basic ${Buffer.from('uzer:1234abcd').toString('base64')}` });
        });

        it('Should use the default api url', async () => {
            const client = new Bitbucket({ password: '1234abcd' });
            const reqMock = jest.fn(async () => getFileResponse());
            request.mockImplementationOnce(reqMock);

            await client.fetchRawFile({ owner: 'joe', repo: 'repo', ref: 'master', path: './some-dir/codefresh.yml' });
            const expectedPath = encodeURIComponent('some-dir/codefresh.yml');
            expect(reqMock).toHaveBeenCalledTimes(1);
            expect(reqMock.mock.calls[0][0].url).toEqual(`https://bitbucket.org/api/2.0/repositories/joe/repo/src/master/${expectedPath}`);
        });

        it('Should use the provided api url', async () => {
            const client = new Bitbucket({ password: '1234abcd', apiURL: 'http://some-url.com' });
            const reqMock = jest.fn(async () => getFileResponse());
            request.mockImplementationOnce(reqMock);

            await client.fetchRawFile({ owner: 'joe', repo: 'repo', ref: 'master', path: './some-dir/codefresh.yml' });
            const expectedPath = encodeURIComponent('some-dir/codefresh.yml');
            expect(reqMock).toHaveBeenCalledTimes(1);
            expect(reqMock.mock.calls[0][0].url).toEqual(`http://some-url.com/api/2.0/repositories/joe/repo/src/master/${expectedPath}`);
        });
    });

    describe('Positive', () => {
        describe('fetchRawFile', () => {
            it('Should fetch the requested file', async () => {
                const client = new Bitbucket({ password: '1234abcd' });
                const reqMock = jest.fn(async () => getFileResponse());
                request.mockImplementationOnce(reqMock);

                const res = await client.fetchRawFile({ owner: 'joe', repo: 'some-repo', ref: 'master', path: 'some-path' });
                expect(reqMock).toHaveBeenCalledTimes(1);
                expect(res).toEqual('some-text');
            });
        });

        describe('getBranch', () => {
            it('Should get the requested branch', async () => {
                const client = new Bitbucket({ password: '1234abcd' });
                const reqMock = jest.fn(async () => getBranchResponse());
                request.mockImplementationOnce(reqMock);

                const branchResponse = getBranchResponse();
                const res = await client.getBranch({ owner: 'joe', repo: 'some-repo', branch: 'master' });
                expect(reqMock).toHaveBeenCalledTimes(1);
                expect(res.id).toEqual(branchResponse.body.name);
                expect(res.name).toEqual(branchResponse.body.name);
                expect(res.commit.commiter_name).toEqual(branchResponse.body.target.author.user.display_name);
                expect(res.commit.message).toEqual(branchResponse.body.target.message);
                expect(res.commit.sha).toEqual(branchResponse.body.target.hash);
                expect(res.commit.url).toEqual(branchResponse.body.target.links.html.href);
            });
        });
    });

    describe('Negative', () => {
        describe('fetchRawFile', () => {
            it('Should fail if the requested file is not found', async () => {
                const client = new Bitbucket({ password: '1234abcd' });
                const reqMock = jest.fn(async () => getFileResponse(404));
                request.mockImplementationOnce(reqMock);

                const opts = { owner: 'joe', repo: 'some-repo', ref: 'master', path: 'some-path' };
                const [err] = await to(client.fetchRawFile(opts));

                expect(err).toBeDefined();
                expect(err.message).toEqual(`Failed to retrieve file: ${JSON.stringify(opts)}, status code: 404`);
            });
        });

        describe('getBranch', () => {
            it('Should fail if the requested branch is not found', async () => {
                const client = new Bitbucket({ password: '1234abcd' });
                const reqMock = jest.fn(async () => getBranchResponse(404));
                request.mockImplementationOnce(reqMock);

                const opts = { owner: 'joe', repo: 'some-repo', branch: 'master' };
                const [err] = await to(client.getBranch(opts));

                expect(err).toBeDefined();
                expect(err.message).toEqual(`Failed to retrieve branch: ${JSON.stringify(opts)}, status code: 404`);
            });
        });
    });
});
