const Github = require('../github').default;
const OctokitMock = require('../__mocks__/@octokit/rest');
const { to } = require('../../helpers');

const getBranchResponse = (status = 200) => {
    return {
        status,
        data: {
            name: 'master',
            commit: {
                sha: '05a7d037f1317877cf1719db1c231a08cc6c7788',
                commit: {
                    author: {
                        name: 'Joe Schmo',
                    },
                    message: 'some commit message',
                    url: 'https://gitlab.com/joeschmoe/codefresh-test/-/commit/05a7d037f1317877cf1719db1c231a08cc6c7788',
                }
            },
        },
    };
};

const getFileResponse = (status = 200) => {
    return {
        status,
        data: {
            content: Buffer.from('some-text').toString('base64'),
        },
    };
};

const getProviderMock = (getBranchFn = getBranchResponse, getFileFn = getFileResponse) => {
    const provider = function () {
        this.repos = {
            getBranch: jest.fn(getBranchFn),
            getContent: jest.fn(getFileFn),
        };
        return this;
    };
    return provider;
};

describe('Github git provider tests', () => {
    describe('Constructor', () => {
        it('Should construct a client', () => {
            const client = new Github({});
            expect(client).toBeInstanceOf(Github);
            expect(client.fetchRawFile).toBeDefined();
            expect(client.getBranch).toBeDefined();
        });

        it('Should use the provided token in the Authorization header', async () => {
            // eslint-disable-next-line prefer-arrow-callback
            const constructorMock = jest.fn(function () { return ({}); });
            OctokitMock.Octokit.mockImplementationOnce(constructorMock);

            // eslint-disable-next-line no-new
            new Github({ password: '1234abcd' });

            expect(constructorMock.mock.calls[0][0].auth).toEqual('1234abcd');
        });

        it('Should use the default api url', async () => {
            // eslint-disable-next-line prefer-arrow-callback
            const constructorMock = jest.fn(function () { return ({}); });
            OctokitMock.Octokit.mockImplementationOnce(constructorMock);

            // eslint-disable-next-line no-new
            new Github({});

            expect(constructorMock.mock.calls[0][0].baseUrl).toEqual('https://api.github.com');
        });

        it('Should use the provided api url', async () => {
            // eslint-disable-next-line prefer-arrow-callback
            const constructorMock = jest.fn(function () { return ({}); });
            OctokitMock.Octokit.mockImplementationOnce(constructorMock);

            // eslint-disable-next-line no-new
            new Github({ apiHost: 'some.github.host' });

            expect(constructorMock.mock.calls[0][0].baseUrl).toEqual('https://some.github.host');
        });

        it('Should not add path prefix if not provided with one', async () => {
            // eslint-disable-next-line prefer-arrow-callback
            const constructorMock = jest.fn(function () { return ({}); });
            OctokitMock.Octokit.mockImplementationOnce(constructorMock);

            // eslint-disable-next-line no-new
            new Github({});

            expect(constructorMock.mock.calls[0][0].baseUrl).toEqual('https://api.github.com');
        });

        it('Should use the provided path prefix', async () => {
            // eslint-disable-next-line prefer-arrow-callback
            const constructorMock = jest.fn(function () { return ({}); });
            OctokitMock.Octokit.mockImplementationOnce(constructorMock);

            // eslint-disable-next-line no-new
            new Github({ apiPathPrefix: '/api/v3', apiHost: 'some-host.com' });

            expect(constructorMock.mock.calls[0][0].baseUrl).toEqual('https://some-host.com/api/v3');
        });
    });

    describe('Positive', () => {
        describe('fetchRawFile', () => {
            it('Should fetch the requested file', async () => {
                OctokitMock.Octokit.mockImplementationOnce(getProviderMock());
                const client = new Github({});
                const res = await client.fetchRawFile({ owner: 'joe', repo: 'some-repo', ref: 'master', path: 'some-path' });
                expect(res).toEqual('some-text');
            });

            it('Should clean file path', async () => {
                const getFileMock = jest.fn(async () => ({
                    status: 200,
                    data: {
                        content: Buffer.from('some-text').toString('base64'),
                    },
                }));
                OctokitMock.Octokit.mockImplementationOnce(getProviderMock(undefined, getFileMock));
                const client = new Github({});
                await client.fetchRawFile({ owner: 'joe', repo: 'some-repo', ref: 'master', path: './dir/some-path' });
                expect(getFileMock.mock.calls[0][0].path).toEqual('dir/some-path');
            });
        });

        describe('getBranch', () => {
            it('Should get the requested branch', async () => {
                OctokitMock.Octokit.mockImplementationOnce(getProviderMock());
                const client = new Github({});

                const branchResponse = getBranchResponse();
                const res = await client.getBranch({ owner: 'joe', repo: 'some-repo', branch: 'master' });
                expect(res.id).toEqual(branchResponse.data.name);
                expect(res.name).toEqual(branchResponse.data.name);
                expect(res.commit.commiter_name).toEqual(branchResponse.data.commit.commit.author.name);
                expect(res.commit.message).toEqual(branchResponse.data.commit.commit.message);
                expect(res.commit.sha).toEqual(branchResponse.data.commit.sha);
                expect(res.commit.url).toEqual(branchResponse.data.commit.commit.url);
            });
        });
    });

    describe('Negative', () => {
        describe('fetchRawFile', () => {
            it('Should fail if the requested file is not found', async () => {
                const err = new Error('some error message');
                err.status = 404;
                OctokitMock.Octokit.mockImplementationOnce(getProviderMock(undefined, async () => { throw err; }));
                const client = new Github({});

                const opts = { owner: 'joe', repo: 'some-repo', ref: 'master', path: 'some-path' };
                const [fetchFileErr] = await to(client.fetchRawFile(opts));

                expect(fetchFileErr).toBeDefined();
                expect(fetchFileErr.message).toEqual(`Failed to retrieve file: ${JSON.stringify(opts)}, status code: 404`);
            });
        });

        describe('getBranch', () => {
            it('Should fail if the requested branch is not found', async () => {
                const err = new Error('some error message');
                err.status = 404;
                OctokitMock.Octokit.mockImplementationOnce(getProviderMock(async () => { throw err; }));
                const client = new Github({});

                const opts = { owner: 'joe', repo: 'some-repo', branch: 'master' };
                const [getBranchErr] = await to(client.getBranch(opts));

                expect(getBranchErr).toBeDefined();
                expect(getBranchErr.message).toEqual(`Failed to retrieve branch: ${JSON.stringify(opts)}, status code: 404`);
            });
        });
    });
});
