jest.mock('request-promise');

const request = require('request-promise');
const Gitlab = require('../gitlab').default;
const { to } = require('../../helpers');

const getBranchResponse = (status = 200) => {
    return {
        statusCode: status,
        body: {
            name: 'master',
            commit: {
                id: '05a7d037f1317877cf1719db1c231a08cc6c7788',
                short_id: '05a7d037',
                message: 'some commit message',
                author_name: 'Joe Schmo',
                web_url: 'https://gitlab.com/joeschmoe/codefresh-test/-/commit/05a7d037f1317877cf1719db1c231a08cc6c7788'
            },
            web_url: 'https://gitlab.com/joeschmo/codefresh-test/-/tree/master',
        },
    };
};

const getFileResponse = (status = 200) => {
    return {
        statusCode: status,
        body: {
            content: Buffer.from('some-text').toString('base64'),
        },
    };
};

describe('Gitlab git provider tests', () => {
    describe('Constructor', () => {
        it('Should construct a client', () => {
            const client = new Gitlab({});
            expect(client).toBeInstanceOf(Gitlab);
            expect(client.fetchRawFile).toBeDefined();
            expect(client.getBranch).toBeDefined();
        });

        it('Should use the provided token in the Authorization header', async () => {
            const client = new Gitlab({ password: '1234abcd' });
            const reqMock = jest.fn(async () => getFileResponse());
            request.mockImplementationOnce(reqMock);

            await client.fetchRawFile({ owner: 'joe', repo: 'some-repo', ref: 'master', path: 'some-path' });
            expect(reqMock).toHaveBeenCalledTimes(1);
            expect(reqMock.mock.calls[0][0].headers).toEqual({ Authorization: 'Bearer 1234abcd' });
        });

        it('Should use the default api url', async () => {
            const client = new Gitlab({ password: '1234abcd' });
            const reqMock = jest.fn(async () => getFileResponse());
            request.mockImplementationOnce(reqMock);

            await client.fetchRawFile({ owner: 'joe', repo: 'repo', ref: 'master', path: './some-dir/codefresh.yml' });
            const expectedPath = encodeURIComponent('some-dir/codefresh.yml');
            expect(reqMock).toHaveBeenCalledTimes(1);
            expect(reqMock.mock.calls[0][0].url).toEqual(`https://gitlab.com/api/v4/projects/joe%2Frepo/repository/files/${expectedPath}`);
            expect(reqMock.mock.calls[0][0].qs.ref).toEqual('master');
        });

        it('Should use the provided api url', async () => {
            const client = new Gitlab({ password: '1234abcd', apiURL: 'http://some-url.com/api/v25' });
            const reqMock = jest.fn(async () => getFileResponse());
            request.mockImplementationOnce(reqMock);

            await client.fetchRawFile({ owner: 'joe', repo: 'repo', ref: 'master', path: './some-dir/codefresh.yml' });
            const expectedPath = encodeURIComponent('some-dir/codefresh.yml');
            expect(reqMock).toHaveBeenCalledTimes(1);
            expect(reqMock.mock.calls[0][0].url).toEqual(`http://some-url.com/api/v25/projects/joe%2Frepo/repository/files/${expectedPath}`);
            expect(reqMock.mock.calls[0][0].qs.ref).toEqual('master');
        });
    });

    describe('Positive', () => {
        describe('fetchRawFile', () => {
            it('Should fetch the requested file', async () => {
                const client = new Gitlab({ password: '1234abcd' });
                const reqMock = jest.fn(async () => getFileResponse());
                request.mockImplementationOnce(reqMock);

                const res = await client.fetchRawFile({ owner: 'joe', repo: 'some-repo', ref: 'master', path: 'some-path' });
                expect(reqMock).toHaveBeenCalledTimes(1);
                expect(res).toEqual('some-text');
            });
        });

        describe('getBranch', () => {
            it('Should get the requested branch', async () => {
                const client = new Gitlab({ password: '1234abcd' });
                const reqMock = jest.fn(async () => getBranchResponse());
                request.mockImplementationOnce(reqMock);

                const branchResponse = getBranchResponse();
                const res = await client.getBranch({ owner: 'joe', repo: 'some-repo', branch: 'master' });
                expect(reqMock).toHaveBeenCalledTimes(1);
                expect(res.id).toEqual(branchResponse.body.name);
                expect(res.name).toEqual(branchResponse.body.name);
                expect(res.commit.commiter_name).toEqual(branchResponse.body.commit.author_name);
                expect(res.commit.message).toEqual(branchResponse.body.commit.message);
                expect(res.commit.sha).toEqual(branchResponse.body.commit.id);
                expect(res.commit.url).toEqual(branchResponse.body.commit.web_url);
            });
        });
    });

    describe('Negative', () => {
        describe('fetchRawFile', () => {
            it('Should fail if the requested file is not found', async () => {
                const client = new Gitlab({ password: '1234abcd' });
                const reqMock = jest.fn(async () => getFileResponse(404));
                request.mockImplementationOnce(reqMock);

                const opts = { owner: 'joe', repo: 'some-repo', ref: 'master', path: 'some-path' };
                const [err] = await to(client.fetchRawFile(opts));

                expect(err).toBeDefined();
                expect(err.message).toEqual(`Failed to retrieve file: ${JSON.stringify(opts)}, status code: 404`);
            });
        });

        describe('getBranch', () => {
            it('Should fail if the requested branch is not found', async () => {
                const client = new Gitlab({ password: '1234abcd' });
                const reqMock = jest.fn(async () => getBranchResponse(404));
                request.mockImplementationOnce(reqMock);

                const opts = { owner: 'joe', repo: 'some-repo', branch: 'master' };
                const [err] = await to(client.getBranch(opts));

                expect(err).toBeDefined();
                expect(err.message).toEqual(`Failed to retrieve branch: ${JSON.stringify(opts)}, status code: 404`);
            });
        });
    });
});
