import _ from 'lodash'
import rp from 'request-promise'
import {
    Provider,
    Branch,
} from './types'

// tslint:disable-next-line: no-var-requires
const CFError = require('cf-errors')

const LIMIT_PER_PAGE = 500
const MAX_PAGES = 20
const ApiVersions = {
    V1: 'rest/api/1.0/',
}

const _cleanEncodedFilePath = (filepath: string): string => {
    const prefix = './'
    const hasPrefix = filepath.indexOf(prefix) === 0
    const path = hasPrefix ? filepath.substring(prefix.length) : filepath
    return encodeURIComponent(path)
}

export default class BitbucketServer implements Provider {
    private baseUrl: string
    private authenticationHeader: { [key: string]: string }
    private timeout: number

    constructor(opts: {
        password: string,
        username?: string,
        type: 'basic' | 'oauth2' | 'PAT',
        apiURL?: string,
        timeout?: number,
    }) {
        if (!opts.apiURL) {
            throw new CFError('Cannot create bitbucket-server provider without api url')
        }
        this.baseUrl = opts.apiURL
        if (!this.baseUrl.endsWith('/')) {
            this.baseUrl = `${this.baseUrl}/`
        }
        this.timeout = opts.timeout || 10000
        if (opts.type === 'oauth2') {
            this.authenticationHeader = { Authorization: `Bearer ${opts.password}`}
        } else {
            this.authenticationHeader = { Authorization: `Basic ${Buffer.from(`${opts.username}:${opts.password}`).toString('base64')}`}
        }
    }

    private async performAPICall(opts: {
        method?: string,
        api: string,
        apiVersion: string,
        headers?: { [key: string]: string },
        qs?: { [key: string]: string },
        json?: boolean,
        noAuth?: boolean,
        data?: any,
        noRefresh?: boolean
    }): Promise<any> {
        const method = opts.method || 'GET'
        const requestHeaders = _.merge(opts.noAuth ? {} : this.authenticationHeader, opts.headers || {})

        const requestOptions = {
            method,
            url: `${this.baseUrl}${opts.apiVersion}${opts.api}`,
            qs: opts.qs || {},
            headers: requestHeaders,
            json: opts.json,
            timeout: this.timeout,
            form: opts.data,
            resolveWithFullResponse: true,
        }

        return rp(requestOptions)
    }

    async fetchRawFile(opts: { owner: string; repo: string; ref: string; path: string }): Promise<string> {
        const lines: string[] = []
        let start = 0
        let page = 0
        let isLastPage = false

        while (!isLastPage && page < MAX_PAGES) {
            const res = await this.performAPICall({
                apiVersion: ApiVersions.V1,
                api: `projects/${opts.owner}/repos/${opts.repo}/browse/${_cleanEncodedFilePath(opts.path)}`,
                json: true,
                qs: {
                    at: opts.ref,
                    raw: 'true',
                    start: '' + start,
                    limit: '' + LIMIT_PER_PAGE,
                }
            })
            if (res.statusCode >= 400) {
                throw new CFError(`Failed to retrieve file: ${JSON.stringify(opts)}, status code: ${res.statusCode}`)
            }

            isLastPage = res.body.isLastPage
            lines.push(..._.map(res.body.lines, (line: any) => line.text as string))
            start += LIMIT_PER_PAGE
            page += 1
        }

        return lines.join('\n')
    }

    async getBranch(opts: { owner: string, repo: string, branch: string }): Promise<Branch> {
        const commitMetadataPath = 'com.atlassian.bitbucket.server.bitbucket-branch:latest-commit-metadata'
        let branch: any
        let start = 0
        let page = 0
        let isLastPage = false

        while (!isLastPage && page < MAX_PAGES) {
            const res = await this.performAPICall({
                apiVersion: ApiVersions.V1,
                api: `projects/${opts.owner}/repos/${opts.repo}/branches`,
                json: true,
                qs: {
                    filterText: opts.branch,
                    details: 'true',
                    start: '' + start,
                    limit: '' + LIMIT_PER_PAGE,
                }
            })
            if (res.statusCode >= 400) {
                throw new CFError(`Failed to retrieve branch: ${JSON.stringify(opts)}, status code: ${res.statusCode}`)
            }

            isLastPage = res.body.isLastPage
            branch = _.find(res.body.values, (b: any) => _.isEqual(opts.branch, b.displayId ? b.displayId : _.last(b.id.split('/'))))
            if (branch) {
                break // found the correct branch
            }
            start += LIMIT_PER_PAGE
            page += 1
        }

        if (!branch) {
            throw new CFError(`Failed to retrieve branch: ${JSON.stringify(opts)}, branch not found`)
        }

        return {
            name: branch.displayId,
            id: branch.id,
            commit: {
                sha: branch.metadata[commitMetadataPath].id,
                commiter_name: branch.metadata[commitMetadataPath].author.name,
                message: branch.metadata[commitMetadataPath].message,
                url: '',
            }
        }
    }
}
