import _ from 'lodash'
import rp from 'request-promise'
import {
    Provider,
    Branch,
} from './types'

// tslint:disable-next-line: no-var-requires
const CFError = require('cf-errors')

const ApiVersions = {
    V1: 'api/1.0/',
    V2: 'api/2.0/',
    OAUTH: 'site/oauth2/access_token',
}

const _cleanEncodedFilePath = (filepath: string): string => {
    const prefix = './'
    const hasPrefix = filepath.indexOf(prefix) === 0
    const path = hasPrefix ? filepath.substring(prefix.length) : filepath
    return encodeURIComponent(path)
}

export default class Bitbucket implements Provider {
    private baseUrl: string
    private authenticationHeader: { [key: string]: string }
    private timeout: number

    constructor(opts: {
        password: string,
        username?: string,
        refreshToken?: string,
        expires?: number,
        type: 'basic' | 'oauth2' | 'PAT',
        apiURL?: string,
        timeout?: number,
    }) {
        this.baseUrl = opts.apiURL || 'https://bitbucket.org/'
        if (!this.baseUrl.endsWith('/')) {
            this.baseUrl = `${this.baseUrl}/`
        }
        this.timeout = opts.timeout || 10000
        if (opts.type === 'oauth2') {
            this.authenticationHeader = { Authorization: `Bearer ${opts.password}`}
        } else {
            this.authenticationHeader = { Authorization: `Basic ${Buffer.from(`${opts.username}:${opts.password}`).toString('base64')}`}
        }
    }

    private async performAPICall(opts: {
        method?: string,
        api: string,
        apiVersion: string,
        headers?: { [key: string]: string },
        qs?: { [key: string]: string },
        json?: boolean,
        noAuth?: boolean,
        data?: any,
        noRefresh?: boolean
    }): Promise<any> {
        const method = opts.method || 'GET'
        const requestHeaders = _.merge(opts.noAuth ? {} : this.authenticationHeader, opts.headers || {})

        const requestOptions = {
            method,
            url: `${this.baseUrl}${opts.apiVersion}${opts.api}`,
            qs: opts.qs || {},
            headers: requestHeaders,
            json: opts.json,
            timeout: this.timeout,
            form: opts.data,
            resolveWithFullResponse: true,
        }

        return rp(requestOptions)
    }

    async fetchRawFile(opts: { owner: string; repo: string; ref: string; path: string }): Promise<string> {
        const res = await this.performAPICall({
            apiVersion: ApiVersions.V2,
            api: `repositories/${opts.owner}/${opts.repo}/src/${opts.ref}/${_cleanEncodedFilePath(opts.path)}`,
        })
        if (res.statusCode >= 400) {
            throw new CFError(`Failed to retrieve file: ${JSON.stringify(opts)}, status code: ${res.statusCode}`)
        }

        return res.body
    }

    async getBranch(opts: { owner: string, repo: string, branch: string }): Promise<Branch> {
        const res = await this.performAPICall({
            apiVersion: ApiVersions.V2,
            api: `repositories/${opts.owner}/${opts.repo}/refs/branches/${opts.branch}`,
            json: true
        })
        if (res.statusCode >= 400) {
            throw new CFError(`Failed to retrieve branch: ${JSON.stringify(opts)}, status code: ${res.statusCode}`)
        }

        return {
            name: res.body.name,
            id: res.body.name,
            commit: {
                sha: res.body.target.hash,
                commiter_name: _.get(res.body, 'target.author.user.display_name', ''),
                message: _.get(res.body, 'target.message', ''),
                url: _.get(res.body, 'target.links.html.href', ''),
            }
        }
    }
}
