import { Octokit } from '@octokit/rest'
import {
    Provider,
    Branch,
} from './types'
import { to } from '../helpers'

// tslint:disable-next-line: no-var-requires
const CFError = require('cf-errors')
// tslint:disable-next-line: no-var-requires
const logger = require('cf-logs').Logger('codefresh:infra:git-providers:github')

const _cleanFilePath = (filepath: string): string => {
    const prefix = './'
    const hasPrefix = filepath.indexOf(prefix) === 0
    const path = hasPrefix ? filepath.substring(prefix.length) : filepath
    return path
}

export default class Github implements Provider {
    private githubClient: Octokit

    constructor(opt: {
        password: string,
        apiPathPrefix?: string,
        apiHost?: string,
        timeout?: number,
    }) {
        const pathPrefix = (opt.apiPathPrefix && opt.apiPathPrefix !== '/') ? opt.apiPathPrefix : ''
        this.githubClient = new Octokit({
            userAgent : 'Codefresh',
            auth : opt.password,
            request:{
                timeout: opt.timeout || 10000,
            },
            timeout: opt.timeout || 10000,
            host: opt.apiHost || 'api.github.com',
            log: logger,
            debug: true,
            baseUrl: `https://${opt.apiHost || 'api.github.com'}${pathPrefix}`
        })
    }

    async fetchRawFile(opt: {owner: string, repo: string, ref: string, path: string}): Promise<string> {
        const {owner, repo, ref, path} = opt
        const [err, res] = await to(this.githubClient.repos.getContent({ owner, repo, path: _cleanFilePath(path), ref }))
        if (err) {
            throw new CFError({
                message: `Failed to retrieve file: ${JSON.stringify(opt)}, status code: ${err.status}`,
                cause: err,
            })
        } else if (res.status >= 400) {
            throw new CFError(`Failed to retrieve file: ${JSON.stringify(opt)}, status code: ${res.status}`)
        }

        const data = res.data
        const base64Content = data.content.replace(/\n/ig, '')
        return Buffer.from(base64Content, 'base64').toString()
    }

    async getBranch(opt: { owner: string; repo: string; branch: string }): Promise<Branch> {
        const [err, res] = await to(this.githubClient.repos.getBranch(opt))
        if (err) {
            throw new CFError({
                message: `Failed to retrieve branch: ${JSON.stringify(opt)}, status code: ${err.status}`,
                cause: err,
            })
        } else if (res.status >= 400) {
            throw new CFError(`Failed to retrieve branch: ${JSON.stringify(opt)}, status code: ${res.status}`)
        }

        return {
            id: res.data.name,
            name: res.data.name,
            commit: {
                sha: res.data.commit.sha,
                commiter_name: res.data.commit.commit.author.name,
                message: res.data.commit.commit.message,
                url: res.data.commit.commit.url,
            }
        }
    }
}
