import _ from 'lodash'
import rp from 'request-promise'
import {
    Provider,
    Branch,
} from './types'

// tslint:disable-next-line: no-var-requires
const CFError = require('cf-errors')

const _encodeProjectPath = (owner: string, repo: string): string => {
    return encodeURIComponent(`${owner}/${repo}`).replace(/\./g, '%2E')
}

const _cleanEncodedFilePath = (filepath: string): string => {
    const prefix = './'
    const hasPrefix = filepath.indexOf(prefix) === 0
    const path = hasPrefix ? filepath.substring(prefix.length) : filepath
    return encodeURIComponent(path)
}

export default class Gitlab implements Provider {
    private baseUrl: string
    private authenticationHeader: { [key: string]: string }
    private timeout: number

    constructor(opts: {
        password: string,
        apiURL?: string,
        timeout?: number,
    }) {
        this.baseUrl = opts.apiURL || 'https://gitlab.com/api/v4/'
        if (!this.baseUrl.endsWith('/')) {
            this.baseUrl = `${this.baseUrl}/`
        }
        this.timeout = opts.timeout || 10000
        // works for both oauth2 and private token authentication methods
        this.authenticationHeader = opts.password ? { Authorization: `Bearer ${opts.password}`} : {}
    }

    private async performAPICall(opts: {
        method?: string,
        api: string,
        headers?: { [key: string]: string },
        qs?: { [key: string]: string },
        json?: boolean,
        noAuth?: boolean,
        data?: any,
    }): Promise<any> {
        const method = opts.method || 'GET'
        const requestHeaders = _.merge(opts.noAuth ? {} : this.authenticationHeader, opts.headers || {})

        const requestOptions = {
            method,
            url: `${this.baseUrl}${opts.api}`,
            qs: opts.qs || {},
            headers: requestHeaders,
            json: opts.json,
            timeout: this.timeout,
            form: opts.data,
            resolveWithFullResponse: true,
        }

        return rp(requestOptions)
    }

    async fetchRawFile(opt: { owner: string; repo: string; ref: string; path: string }): Promise<string> {
        const res = await this.performAPICall({
            api: `projects/${_encodeProjectPath(opt.owner, opt.repo)}/repository/files/${_cleanEncodedFilePath(opt.path)}`,
            qs: { ref: opt.ref },
            json: true,
        })
        if (res.statusCode >= 400) {
            throw new CFError(`Failed to retrieve file: ${JSON.stringify(opt)}, status code: ${res.statusCode}`)
        }

        const base64Content = res.body.content.replace(/\n/ig, '')
        return Buffer.from(base64Content, 'base64').toString()
    }

    async getBranch(opt: { owner: string, repo: string, branch: string }): Promise<Branch> {
        const res = await this.performAPICall({
            api: `projects/${_encodeProjectPath(opt.owner, opt.repo)}/repository/branches/${opt.branch}`,
            json: true
        })
        if (res.statusCode >= 400) {
            throw new CFError(`Failed to retrieve branch: ${JSON.stringify(opt)}, status code: ${res.statusCode}`)
        }

        return {
            name: res.body.name,
            id: res.body.name,
            commit: {
                sha: res.body.commit.id,
                commiter_name: res.body.commit.author_name,
                message: res.body.commit.message,
                url: res.body.commit.web_url,
            }
        }
    }
}
